#viewer-container {
    width: 100%;
    max-width: 100%;
    margin: 0;
    padding: 0;
    box-sizing: border-box;
    display: flex;
}

#controls-container {
    padding: 15px;
    box-sizing: border-box;
    overflow-y: auto;
    max-height: 100vh;
}

#plot-container {
    padding: 15px;
    box-sizing: border-box;
    display: flex;
    align-items: center;
    justify-content: center;
}

#plot-container img {
    max-width: 100%;
    height: auto;
}

.system-controls {
    margin: 10px 0px;
    padding: 10px;
    background-color: #ffffff;
    border: 1px solid #e5e7eb;
    border-radius: 4px;
}

.parameter-controls {
    padding: 10px;
    background-color: #ffffff;
    border: 1px solid #e5e7eb;
    border-radius: 4px;
}

.section-header {
    margin-bottom: 15px;
    font-size: 16px;
}